(function (global) {
    "use strict";

    global.FRCC = global.FRCC || {};
    global.FRCC.Portal = global.FRCC.Portal || {};
    global.FRCC.Portal.Accessibility = (function () {
        const MIN_LOOKUP_IFRAME_HEIGHT = 600;

        /**
         * Global set of accessibility modifications to be made on the portal
         *
         * This will be executed from the EYGA Header Web Template so it is run on every page
         */
        var loadGlobalAccessibilityModifications = async function (configuration) {
            insertSpanElements(configuration.targetingObject);
            improveHeaderAccessibility();
            updateNativeTables();
            updateOfflineNotification();
            updateRestrictedPages(configuration?.portalTitle);
            insertDescriptionBelowTable();
            insertAriaDescribedBy();
            helpTextStyle();
            addHelpPopoverForMultiselect();
            makeTablesHorizontallyScrollable();
            addAutoCompleteAttribute();
            applyTablePaddingActionMenuOffset();
            setTooltipPlacement();
            _fixErrorAlertLists();
            _replaceApplicableLabelsWithLegends();
        };

        /**
         * @param {object} targetingObject - key/value object containing selectors:insertMethod
         *
         * @example
         * insertSpanElements({
         *  'select': 'after',
         *  'input[type="radio"]': 'after',
         *  '.boolean-radio > label': 'prepend'
         * })
         *
         * This is an accessbility-specific function that was needed due to many fields generating without proper
         *      icon identification on entityforms within the portal.
         */
        var insertSpanElements = function (targetingObject) {
            // Cache default jQuery selector
            var $ = jQuery;

            // Native subgrids are generated within iframe elements on entity forms
            // Attaching a handler that will call the function to insert span elements when the iframe has completed its loading
            $("iframe[data-page]").on("load", function (e) {
                try {
                    _insertSpanElementsHelper(targetingObject, e.currentTarget.contentWindow.$);
                } catch (error) {
                    handleError(error);
                }
            });

            _insertSpanElementsHelper(targetingObject, $);
        };

        var _insertSpanElementsHelper = function (targetingObject, $) {
            if (typeof $ === "undefined" || $ === null) {
                return;
            }

            const spanElementHtml = '<span class="ui"></span>';

            for (const property in targetingObject) {
                var selector = property;
                var insertMethod = targetingObject[property];
                switch (insertMethod) {
                    case "after":
                        $(selector).after(spanElementHtml);
                        break;
                    case "prepend":
                        $(selector).prepend(spanElementHtml);
                        break;
                    default:
                        break;
                }
            }
        };

        var resetPageTabbing = function () {
            $("body").focus();
            return;
        };

        var addHelpPopoverForMultiselectHelper = function ($, window) {
            if (typeof $ === "undefined" || $ === null) {
                return;
            }

            var _document = window;
            var observer = new MutationObserver(function () {
                if ($(_document).find(".msos-selecteditems").length) {
                    var multiselectSelectedOptions = $(".msos-selecteditems");
                    var popoverMessage =
                        "Use the arrow keys to navigate through the list of currently selected items. Press delete or backspace to deselect an item.";

                    var popover = `<div class="popover top eyga-msos-popover" aria-hidden="true"><div class="arrow"></div><div class="popover-content">${popoverMessage}</div></div>`;

                    if (multiselectSelectedOptions.length) {
                        multiselectSelectedOptions.after(popover);
                    }
                    observer.disconnect();
                }
            });
            observer.observe(_document, {
                childList: true,
                subtree: true,
            });
        };

        var addHelpPopoverForMultiselect = function () {
            // Cache default jQuery selector
            var $ = jQuery;

            $("iframe[data-page]").on("load", function (e) {
                try {
                    addHelpPopoverForMultiselectHelper(e.currentTarget.contentWindow.$, e.currentTarget.contentWindow.document);
                } catch (error) {
                    handleError(error);
                }
            });

            addHelpPopoverForMultiselectHelper($, document);
        };

        var insertDescriptionBelowTable = function () {
            $(".description.below").find("table").attr("role", "presentation");

            // Native subgrids are generated within iframe elements on entity forms
            // Attaching a handler that will call the function to insert span elements when the iframe has completed its loading
            $("iframe[data-page]").on("load", function (e) {
                try {
                    var $ = e.currentTarget.contentWindow.$;
                    if ($) {
                        $("#EntityFormControl_EntityFormView").find(".description.below").find("table").attr("role", "presentation");
                    }
                } catch (error) {
                    handleError(error);
                }
            });
        };

        /**
         * This function adds an aria-describedby attribute to input elements that is set to the associated
         * label's id
         */
        var insertAriaDescribedBy = function () {
            var $ = jQuery;
            $("iframe[data-page]").on("load", function (e) {
                try {
                    setTimeout(function () {
                        insertAriaDescribedByHelper(e.currentTarget.contentWindow.$);
                    }, 2000);
                } catch (error) {
                    handleError(error);
                }
            });
            setTimeout(function () {
                insertAriaDescribedByHelper($);
            }, 2000);
        };
        /**
         * Helper function that holds logic
         */
        var insertAriaDescribedByHelper = function ($) {
            if (!$) {
                return;
            }

            $("input").each(function () {
                //check here if there is a radio-cell in the parent of info req
                var ifHidden = $(this).attr("type");
                var controlClass = $(this).closest("div.control");
                var isHelpText = controlClass.next().hasClass("description below");
                var infoClass = controlClass.prev();
                if (ifHidden !== "hidden" && isHelpText) {
                    if (infoClass.parent().hasClass("boolean-radio-cell")) {
                        var associatedLabelId = infoClass.children(":first").attr("id");
                        infoClass.parent().attr("aria-describedby", associatedLabelId + "_help");
                        controlClass.next().attr("id", associatedLabelId + "_help");
                    } else {
                        var associatedLabelId = infoClass.children(":first").attr("id");
                        $(this).attr("aria-describedby", associatedLabelId + "_help");
                        controlClass.next().attr("id", associatedLabelId + "_help");
                    }
                }
            });

            $("textarea").each(function () {
                var ifHidden = $(this).attr("type");
                var controlClass = $(this).closest("div.control");
                var infoClass = controlClass.prev();
                var isHelpText = controlClass.next().hasClass("description below");
                if (ifHidden !== "hidden" && isHelpText) {
                    var associatedLabelId = infoClass.children(":first").attr("id");
                    $(this).attr("aria-describedby", associatedLabelId + "_help");
                    controlClass.next().attr("id", associatedLabelId + "_help");
                }
            });

            $("select").each(function () {
                var ifHidden = $(this).attr("type");
                var controlClass = $(this).closest("div.control");
                var infoClass = controlClass.prev();
                var isHelpText = controlClass.next().hasClass("description below");
                if (ifHidden !== "hidden" && isHelpText) {
                    var associatedLabelId = infoClass.children(":first").attr("id");
                    $(this).attr("aria-describedby", associatedLabelId + "_help");
                    controlClass.next().attr("id", associatedLabelId + "_help");
                }
            });
        };

        /**
         * Removes in line styling that interfered with help text readability
         */
        var helpTextStyle = function () {
            var $ = jQuery;
            $("iframe[data-page]").on("load", function (e) {
                try {
                    helpTextStyleHelper(e.currentTarget.contentWindow.$);
                } catch (error) {
                    handleError(error);
                }
            });
            helpTextStyleHelper($);
        };

        var helpTextStyleHelper = function ($) {
            if (!$) {
                return;
            }

            $(".description").each(function () {
                $(this).children(":first").removeAttr("style");
                $(this).children().removeAttr("style");
            });
            $("div.description span").removeAttr("style");
        };

        /**
         * This function improves keyboard accessibility for the main navigation header.
         *
         * If the Escape key is pressed when the focus is within he header and a submenu is open, it removes the
         *      "in" class from the dropdown menu to close it.
         *
         * If the Up or Down keys are pressed and the focus is within the context of the navbar-links menu, it will
         *      mimic tab and shift-tab functionality
         */
        var improveHeaderAccessibility = function () {
            var $nav = $('nav[aria-label="main"]');
            var $navbarToggle = $(".navbar-toggle");
            var $webLinks = $(".weblinks");
            var $weblinkDropdown = $(".weblink.dropdown");

            document.addEventListener("keydown", function (e) {
                if (e.code === "Escape" && $nav.has(e.target).length && $nav.find('[aria-expanded="true"]').length) {
                    $("#navbar").removeClass("in");
                }

                if (
                    (e.code === "ArrowDown" || e.code === "ArrowUp") &&
                    ($navbarToggle[0] === e.target || $webLinks.has(e.target).length) &&
                    !$weblinkDropdown.has(e.target).length &&
                    $nav.find('[aria-expanded="true"]').length
                ) {
                    var currentElement = $(document.activeElement);
                    var nextElement = $(":tabbable").eq($(":tabbable").index(currentElement) + (e.code === "ArrowDown" ? 1 : -1));
                    if (!nextElement.hasClass("navbar-links") && e.code === "ArrowDown") {
                        e.preventDefault();
                        return;
                    }
                    nextElement.focus();
                    e.preventDefault();
                }
            });
        };

        /**
         * This function updates the classes used for native tables loaded onto the portal
         *
         * A MutationObserver is required because the native tables are loaded dynamically after
         *      the page/document is rendered
         *
         * table-fluid causes screen readers to not be able to find the table properly and needs to be replaced
         *
         * This function will replace the table-fluid class with table-mobile: a custom class that copies
         *      the styles used in table-fluid
         */
        var updateNativeTables = function () {
            // Native subgrids are generated within iframe elements on entity forms
            // Attaching a handler that will call the function to create a mutation observer when the iframe has finished loading
            $("iframe[data-page]").on("load", function (e) {
                try {
                    _registerMutationObserver(
                        e.currentTarget.contentWindow.document,
                        null,
                        false,
                        { childList: true, subtree: true },
                        _updateNativeTablesHelper
                    );
                } catch (error) {
                    handleError(error);
                }
            });

            _registerMutationObserver(document, "main", false, { childList: true, subtree: true }, _updateNativeTablesHelper);

            _registerMutationObserver(
                document,
                ".view-pagination:not(.core-pagination-wrapper)",
                false,
                { childList: true, subtree: true },
                _updateNativePaginationHelper
            );
        };

        var makeTablesHorizontallyScrollable = function () {
            $("iframe[data-page]").on("load", function (e) {
                try {
                    makeTablesHorizontallyScrollableHelper(e.currentTarget.contentWindow.document);
                } catch (error) {
                    handleError(error);
                }
            });

            makeTablesHorizontallyScrollableHelper(document);
        };

        var makeTablesHorizontallyScrollableHelper = function (documentContext) {
            const tables = documentContext.querySelectorAll(".table");

            tables.forEach((table) => {
                var scrollWrapper = document.createElement("div");
                scrollWrapper.classList.add("table-scroll-wrapper");
                table.parentNode.insertBefore(scrollWrapper, table);
                scrollWrapper.appendChild(table);
            });
        };

        var _updateNativePaginationHelper = function (mutationsList) {
            mutationsList.forEach((mutation) => {
                if (mutation.target.classList.contains("view-pagination") && mutation.addedNodes.length > 0) {
                    var nativePaginationWrapper = $(".view-pagination:not(.core-pagination-wrapper)");

                    nativePaginationWrapper.find("ul.pagination").wrap("<nav aria-label='pagination'></nav>");

                    nativePaginationWrapper.find("a").each(function () {
                        $(this).removeAttr("role");

                        var ariaLabel = $(this).attr("aria-label");

                        if (ariaLabel == "Previous page" || ariaLabel == "Next page") {
                            var newTitle = $(this).attr("aria-label");
                            $(this).attr("title", newTitle);
                            $(this).removeAttr("data-toggle");
                            $(this).contents().wrap("<span aria-hidden='true'></span>");
                        } else if ($(this).attr("data-page") == "..") {
                            $(this).contents().wrap("<span aria-hidden='true'></span>");
                            $(this).attr("aria-label", "ellipsis indicating non-visible pages");
                        } else if (!$(this).parent().hasClass("active")) {
                            //if it's a number but not active
                            $(this).attr("aria-current", "false");
                        } else {
                            //number that is active
                            var pageNum = $(this).attr("data-page");
                            $(this).attr("aria-label", `page ${pageNum}`);
                        }

                        if ($(this).parent().hasClass("disabled")) {
                            $(this).removeAttr("href");
                            $(this).attr("aria-disabled", "true");
                        } else {
                            $(this).attr("href", "javascript:void(0)");
                        }
                    });
                }
            });
        };

        var _updateNativeTablesHelper = function (mutationsList) {
            mutationsList.forEach((mutation) => {
                mutation.addedNodes.forEach((node) => {
                    if (node.nodeName.toLowerCase() === "table") {
                        node.removeAttribute("aria-relevant"); // Remove aria-relevant="additions" - not supported by most browsers

                        const classNames = node.classList;
                        if (classNames.contains("table-hover") && !classNames.contains("table-custom")) {
                            node.querySelector("th").style.width = "5%";
                        }

                        classNames.forEach((e) => {
                            if (!e.contains("table-fluid")) {
                                return;
                            }

                            var n = e.replace("fluid", "mobile");
                            classNames.replace(e, n);
                        });
                    } else if (node.nodeName.toLowerCase() === "tbody") {
                        const parentNode = node.parentNode;

                        const tdNodes = node.querySelectorAll("td");
                        tdNodes.forEach((td) => {
                            td.classList.add("text-wrap");
                            if (td.textContent === ', ,  ' || !(td.textContent || td.children.length)) {
                                td.textContent = '—';
                                let spanEle = document.createElement('span');
                                spanEle.classList = 'sr-only';
                                spanEle.textContent = 'no data';
                                td.append(spanEle);
                            }
                        });
                    }
                });
            });
        };

        /**
         * Set of accessibility modifications to be made on the portal against the native offlineNotificationBar
         */
        var updateOfflineNotification = function () {
            $("#offlineNotificationBar").attr("role", "alert");
            $("#web").attr("aria-hidden", "true");
            $("#close").attr({
                role: "button",
                tabindex: "0",
                style: "top:0.8rem",
            });
            $("#close").on("keypress", function (e) {
                if (e.key !== "Enter") {
                    return;
                }

                $(this).click();
            });
        };

        /**
         * Set of accessibility modifications to be made on the portal's "Restricted Pages"
         *
         * Restricted Pages are web pages where the HTML cannot be edited via a Web Template such as SignIn, Register, Redeem Invitation
         */
        var updateRestrictedPages = function (portalTitle) {
            const webpageTitle = window?.location?.pathname;
            if (!webpageTitle) {
                return;
            }

            const restrictedPages = ["/signin", "/register", "/account/login/register"];
            var adjustedWebPageTitle = webpageTitle.toLowerCase().replace(/\/$/, "");
            if (!restrictedPages.includes(adjustedWebPageTitle)) {
                return;
            }

            _updatePageTitles(portalTitle);
            _updatePageStructure(webpageTitle);
            _updateRestrictedPageAccessibility();
        };

        /**
         * Updates the restricted page's document.title attribute to be consistent with formatting across the Portal
         */
        var _updatePageTitles = function (portalTitle) {
            if (!portalTitle) {
                return;
            }

            _registerMutationObserver(document, "title", false, { childList: true, subtree: true }, (mutationsList) => {
                _updatePageTitle(mutationsList, portalTitle);
            });
        };

        var _updatePageTitle = function (mutationsList, portalTitle) {
            const documentTitle = document.title;
            const parts = documentTitle.split("-");
            const pageTitle = parts.slice(-1);
            if (pageTitle.length === 0) {
                return;
            }

            if (documentTitle === `${portalTitle} - ${pageTitle[0].trim()}`) {
                return;
            }

            document.title = `${portalTitle} - ${pageTitle[0]}`;
        };

        /**
         * Updates the restricted page's HTML structure to be consistent with structures used on other Web Pages throughout
         *      the portal
         */
        var _updatePageStructure = function (webpageTitle) {
            const pageTitle = webpageTitle.replace(/\//g, "").toLowerCase().trim();

            $("#content").addClass("body-content");
            $("#content").unwrap();
            $("#content").wrap(`<main id="content-container" class="page-content page-${pageTitle}" role="main"></main>`);
            $("#mainContent").removeClass("page-content");
            $("#mainContent").addClass("container");
            $(".nav.nav-tabs.nav-account").prependTo("#mainContent");
        };

        /**
         * Updates the restricted page's accessibility by removing redundant attributes and modifying validation alerts
         */
        var _updateRestrictedPageAccessibility = function () {
            _registerMutationObserver(document, ".form-horizontal", true, { childList: true, subtree: true }, _modifyRestrictPageValidationAlerts);

            $(".nav.nav-tabs.nav-account > li > a").attr("tabindex", "0");
            $(".form-horizontal div.alert-danger").attr({
                role: "alert",
                "aria-live": "assertive",
            });
            $("*[title]").removeAttr("aria-label");
            $("*[title]").removeAttr("title");
        };

        var _modifyRestrictPageValidationAlerts = function (mutationsList) {
            $(".form-horizontal div.alert-danger li:visible a").each(function (index, ele) {
                var $ele = $(ele);
                var id = $ele.attr("href");
                $(id).closest(".form-group").addClass("has-error");
            });
        };

        var _registerMutationObserver = function (_document, querySelector, multiple, config, handler) {
            const observer = new MutationObserver(handler);

            const targetNodes = multiple
                ? querySelector
                    ? _document.querySelectorAll(querySelector)
                    : [_document.body]
                : querySelector
                    ? _document.querySelector(querySelector) !== null
                        ? [_document.querySelector(querySelector)]
                        : [_document.body]
                    : [_document.body];

            targetNodes.forEach((node) => observer.observe(node, config));
        };

        var _registerResizeObserver = function (_document, querySelector, multiple, handler) {
            const observer = new ResizeObserver(handler);

            const targetNodes = multiple
                ? querySelector
                    ? _document.querySelectorAll(querySelector)
                    : [_document.body]
                : querySelector
                    ? _document.querySelector(querySelector) !== null
                        ? [_document.querySelector(querySelector)]
                        : [_document.body]
                    : [_document.body];

            targetNodes.forEach((node) => observer.observe(node));
        };

        var addAutoCompleteAttribute = function () {
            var $ = jQuery;
            $("iframe[data-page]").on("load", function (e) {
                try {
                    addAutoCompleteAttributeHelper(e.currentTarget.contentWindow.$);
                } catch (error) {
                    handleError(error);
                }
            });
            addAutoCompleteAttributeHelper($);
        };

        /**
         * This adds required autocomplete attribute with correct token to input elements on an application
         */
        var addAutoCompleteAttributeHelper = function ($) {
            if (typeof $ === "undefined" || $ === null) {
                return;
            }

            var allInputs = $("input, select");
            allInputs.each(function () {
                const inputId = $(this).attr("id");
                if (inputId?.includes("address") && inputId?.includes("street") && inputId?.includes("1")) {
                    $(this).attr("autocomplete", "address-line1");
                } else if (inputId?.includes("address") && inputId?.includes("street") && inputId?.includes("2")) {
                    $(this).attr("autocomplete", "address-line2");
                } else if (inputId?.indexOf("city") !== -1) {
                    $(this).attr("autocomplete", "address-level2");
                } else if (inputId?.indexOf("state") !== -1) {
                    $(this).attr("autocomplete", "address-level1");
                } else if (inputId?.indexOf("postalcode") !== -1 || inputId.indexOf("zip") !== -1) {
                    $(this).attr("autocomplete", "postal-code");
                } else if (inputId?.indexOf("phone") !== -1) {
                    $(this).attr("autocomplete", "tel");
                } else if (inputId?.indexOf("email") !== -1) {
                    $(this).attr("autocomplete", "email");
                }
            });
        };

        /* Apply a padding and negative margin to table scroller so table doesnt clip action menu
         * CSS will pull in document CSS var to use as the value on non-mobile layouts
         */
        var applyTablePaddingActionMenuOffset = function () {
            $("table").on("show.bs.dropdown", function ({ target }) {
                /* If table is part of the application, dont do this
                 * Application tables use fixed position to position dropdown
                 */
                if ($("table").closest(".view-grid").length) {
                    return;
                }

                const dropdownMenu = target.querySelector(".dropdown-menu");
                const dropdownMenuHeight = $(dropdownMenu).outerHeight();
                const heightAllowance = 5; // Value to add to the height so its has some wiggle room

                document.documentElement.style.setProperty("--dropdownMenuHeight", `${dropdownMenuHeight + heightAllowance}px`);
            });
        };

        /**
         * Application-specific set of accessibility modifications to be made on the portal
         *
         * This will be executed from the EYGA Templates: Application Create/Edit/View Web Templates
         */
        var loadApplicationAccessibilityModifications = async function (isReadonly) {
            _removeRedundantAttributes();
            _modifyValidationAlerts();
            _modifyTableAlerts();
            _updateTableResponsive();
            _fixTableActionsHeader();
            _addPlaceholderText();
            _modifyProgressIndicatorModal();
            _removeRoleDescriptions();
            _modifyRequiredFieldLabels();
            _addApplicationLoadingIndicators();
            _modifyNativeModalIFrames();
            _hideInstructionsIfEmpty();
            _reverseRadioBooleanCells();
            await _modifyMultiStepNavigationElements();
            _checkForEditElementAndUncheckRadios();
            _uncheckDefaultedRadioButtons();
            _insertHelpSpan();
            _removeFormElementDisabledAttributes();
            _removeAriaLabels();
            _removeAriaReadOnlyOnTables();
            _applyDatePickerAriaAttributes();
            _addTableCaptions();
            _addDocumentDownloadAccessibleText();
            _updateTableActionIcons();
            _modifyMultilineTextRowHeight();
            _applyCurrencyFieldPostProcessing(isReadonly);
            _checkForUnsavedChangesPrevious();
            _fixNativeNotificationHTMLAccessibility();
            _removePlaceholderText();
            _addApplicationFormSkipLink();
            _removeCalendarsForReadonlyDateTimePickers(isReadonly);
            _setFocus();
        };

        /**
         * Removes "aria-label" and "title" attributes from required fields on entityforms within the portal.
         *
         * This is an accessbility-specific function that was needed due to required fields generating
         *      with additional attributes that non-required fields were not generating with.
         * The aria-label and title attributes were causing screen readers to be read back the same information
         *      to the user, causing confusing and redundancy.
         */
        var _removeRedundantAttributes = function () {
            var $ = jQuery;
            $("iframe[data-page], iframe[data-formname]").on("load", function (e) {
                try {
                    _removeRedundantAttributesHelper(e.currentTarget.contentWindow.$);
                } catch (error) {
                    handleError(error);
                }
            });

            _removeRedundantAttributesHelper($);
        };

        var _removeRedundantAttributesHelper = function ($) {
            if (typeof $ === "undefined" || $ === null) {
                return;
            }

            const $entityForm = $(".crmEntityFormView");

            $entityForm.find('.required + .control > [aria-required="true"]').removeAttr("aria-label title");
            $entityForm.find(".required + .control > select").removeAttr("title");
            $entityForm.find('input[type="text"][required]').removeAttr("required").attr("aria-required", "true");

            $entityForm.find("input[readonly], label[readonly]").removeAttr("aria-label");
            $entityForm.find("input[readonly], label[readonly]").removeAttr("title");
            $entityForm.find("label[readonly]").removeAttr("readonly");

            $entityForm.find("input[aria-roledescription], select[aria-roledescription], textarea[aria-roledescription]").removeAttr("aria-roledescription");

            $entityForm.find("fieldset[aria-label], form[aria-label]").removeAttr("aria-label");
            $entityForm.find("fieldset[aria-labelledby], form[aria-labelledby]").removeAttr("aria-labelledby");

            // Remove aria-hidden from readonly fields that have no data (which display an emdash), and add an aria-label that states No data for screenreaders
            $entityForm.find("div.text-muted[aria-hidden='true']").removeAttr("aria-hidden").attr("aria-label", "No data");

            // disabled inputs either have readonly, disabled, aria-disabled, or aria-readonly (or any combination of the 4)
            // we wish to make these consistent by removing all of them except readonly
            let disabledInputs = $entityForm.find("input[disabled], input[aria-disabled], input[aria-readonly]");
            disabledInputs.prop("readonly", true);
            disabledInputs.removeAttr("disabled").removeAttr("aria-disabled").removeAttr("aria-readonly");

            // entityForm may be pointing to <div id="WebFormPanel" role="form"> which we want to remove aria-label from
            if ($entityForm.length > 0 && $entityForm[0].tagName === "DIV" && $entityForm[0].role === "form") {
                $entityForm.removeAttr("aria-label");
            }
        };

        /**
         * Adds role="alert" to the validation summary elements on entityforms within the portal and ensures that
         * fields that are listed in the validation summary are given the 'has-error' class. Also deletes all divs 
         * with class "validators" as we do not want to show the native asterisk on fields that are incorrect.
         *
         * This is an accessibility-specific function that was needed due to the validation areas on the form
         *      not having the proper role to inform screen reader users of the intended function of the element(s)
         *      on their screen.
         */
        var _modifyValidationAlerts = function () {
            var $ = jQuery;

            $("iframe[data-page], iframe[data-formname]").on("load", function (e) {
                try {
                    let iframe$ = e.currentTarget.contentWindow.$;
                    _modifyValidationAlertsHelper(e.currentTarget.contentWindow.document, iframe$);
                    _addHasErrorClassToControlsWithInvalidInput(iframe$);
                    if (iframe$) {
                        iframe$(".validators").remove();
                    }
                } catch (error) {
                    handleError(error);
                }
            });

            _modifyValidationAlertsHelper(document, $);
            _addHasErrorClassToControlsWithInvalidInput($);
            $(".validators").remove();
        };

        var _modifyValidationAlertsHelper = function (documentContext, $) {
            if (!$) {
                return;
            }

            var validationBlockSelector = ".validation-summary.alert.alert-error.alert-danger.alert-block";
            var validationBlock = $(validationBlockSelector);
            validationBlock.attr("role", "alert");

            _registerMutationObserver(documentContext, validationBlockSelector, true, { childList: true, subtree: true }, _addErrorClasses);

            $('input[value="x"]').val("×");
        };

        var _modifyMultilineTextRowHeight = function () {
            var $ = jQuery;
            $("iframe[data-page],iframe[data-formname]").on("load", function (e) {
                try {
                    _modifyMultilineTextRowHeightHelper(e.currentTarget.contentWindow.$);
                    _resizeQuickviewFormHeight();
                } catch (error) {
                    handleError(error);
                }
            });
            _modifyMultilineTextRowHeightHelper($);
        };

        var _modifyMultilineTextRowHeightHelper = function ($) {
            if (!$) {
                return;
            }

            $("textarea.textarea,textarea.form-control").attr("rows", 10);
        };

        const _resizeQuickviewFormHeight = function () {
            const lookupButtons = $("iframe[data-formname]").find("button.btn.btn-default.launchentitylookup > span.fa.fa-search");
            const currentInnerHeight = $("iframe[data-formname]").contents().find(".container.quickform").height();

            if (lookupButtons.length && currentInnerHeight < MIN_LOOKUP_IFRAME_HEIGHT) {
                $("iframe[data-formname]").height(MIN_LOOKUP_IFRAME_HEIGHT);
            } else {
                $("iframe[data-formname]").height(currentInnerHeight);
            }
        };

        var _addErrorClasses = function (mutationsList, observer) {
            mutationsList.forEach((mutation) => {
                const $parentForm = $(mutation.target).closest(".entity-form");

                if (mutation.addedNodes.length > 0) {
                    $(mutation.target).closest(".entity-form").find(".has-error").removeClass("has-error");
                    mutation.target.setAttribute("role", "alert");
                    mutation.target.setAttribute("aria-live", "assertive");
    
                    mutation.addedNodes.forEach((node) => {
                        if (node.nodeName.toLowerCase() !== "ul") {
                            return;
                        }
    
                        node.removeAttribute("role");
                        // iterate backword through the list as we may be removing childNodes
                        for (var i = node.childNodes.length - 1; i >= 0; i--) {
                            let childNode = node.childNodes[i];
                            _addHasErrorClassHelper($, $parentForm, childNode.childNodes[0]);
                        }
                    });
                }
            });
        };

        const _addHasErrorClassToControlsWithInvalidInput = function ($) {
            if (!$) {
                return;
            }

            let errorNotificationLinks = $(".validation-summary").find("a");
            const $parentForm = $(errorNotificationLinks).closest(".entity-form");
            errorNotificationLinks.each((index, linkElement) => {
                _addHasErrorClassHelper($, $parentForm, linkElement);
            });
        }

        /**
         * Add has-error class to the parent td of the input control with
         * invalid input if the input control is not readonly
         *
         * @param $ JQuery object for the window where the form is located
         * @param $parentForm The parent form of the input control
         * @param validationLinkElement The validation link element from the validation summary
         */
        const _addHasErrorClassHelper = function ($, $parentForm, validationLinkElement) {
            var href = validationLinkElement.getAttribute("href").replace("#", "");

            // to ensure that we correctly add has-error to project dropdown instead of to the hidden lookup control
            // we also want to make sure the link to the project field in the validation summary is correct and not pointing to hidden lookup control
            if (href === "eyfrcc_applicationprojectid") {
                validationLinkElement.setAttribute("href", "#project-fake-dropdown");
                href = "project-fake-dropdown";
            }

            if (href.includes("_label")) {
                let formControlCell = $parentForm.find(`#${href}`).parents("td");
                // remove <li> for this link if the control itself is not visible
                if (!formControlCell.is(":visible")) {
                    $(validationLinkElement).parent().remove();
                } else {
                    if (!_checkIfFormControlIsReadonly(formControlCell)) {
                        formControlCell.addClass("has-error");
                    }
                }
            } else if (href.includes("WebResource_FileUpload_")) {
                document.getElementById(href).parentElement.previousElementSibling.classList.add("has-error");
            } else {
                let formControlCell = $(`#${href}`).parents("td");
                
                // remove <li> for this link if the control itself is not visible
                if (!formControlCell.is(":visible")) {
                    $(validationLinkElement).parent().remove();
                } else {
                    if (!_checkIfFormControlIsReadonly(formControlCell)) {
                        formControlCell.addClass("has-error");
                    }
                }
            }
        }

        /**
         * Check if the form control is readonly
         *
         * @param formControlCell The form control
         * @returns {boolean} true if the form control is readonly, false otherwise
         */
        const _checkIfFormControlIsReadonly = (formControlCell) => {
            if (formControlCell.hasClass("lookup")) {
                // lookup controls need to be managed slightly differently
                // if modal exists, this field is not readonly. 
                let modal = formControlCell.find(".lookup-modal");
                let isLookupModalPresent = modal && modal.length > 0;

                let isFakeProjectDropdown = formControlCell.find("#project-fake-dropdown");
                let isFakeProjectDropdownPresent = isFakeProjectDropdown && isFakeProjectDropdown.length > 0;

                // if the modal is not there, and it's not the fake project dropdown, it is readonly
                return !isLookupModalPresent && !isFakeProjectDropdownPresent;
            }
            else {
                let inputControl = formControlCell.find("input,select");
                return inputControl && inputControl.length > 0 && inputControl[0].hasAttribute("readonly")
            }
        }

        var _modifyTableAlerts = function () {
            $(".view-empty").removeAttr("role");
            $(".view-access-denied").attr("role", "alert");
            $(".view-error").attr("role", "alert");
        };

        /**
         * Adds placeholder text to select controls on entityforms within the portal
         *
         * This is an accessibility-specific function that was needed due to selects natively showing no placeholder
         *      text when no option was selected
         */
        var _addPlaceholderText = function () {
            $(".form-control.picklist option:first-child").text("Please select an option");
        };

        /**
         * Adds aria-atomic="true" and aria-live="assertive" to parent-level element of the progressIndicator element
         *
         * Since the Progress Indicator is loaded via a Content Snippet, it comes with extra baggage - 2 parent divs
         *      are loaded along with the HTML content of the Content Snippet. In order to correctly mark the element
         *      to screen readers as a "live" area, we have to do it at the parent-level
         */
        var _modifyProgressIndicatorModal = function () {
            var progressIndicator = $("#progressIndicator");
            if (progressIndicator.length === 0) {
                return;
            }

            progressIndicator.parent().parent().attr({ "aria-atomic": "true", "aria-live": "assertive" });
        };

        /**
         * Removes "aria-roledescription" from all inputs on entityforms within the portal.
         *
         * This is an accessbility-specific function that was needed due the aria-roledescription attribute
         *      overriding the selected attribute and reading the content of aria-roledescription to screen readers.
         */
        var _removeRoleDescriptions = function () {
            $(".crmEntityForm [aria-roledescription]").removeAttr("aria-roledescription");
        };

        /**
         * Adds <span class="sr-only">*</span> to all required input labels within the portal.
         *
         * This is an accessbility-specific function that was needed due to labels for required inputs
         *      not having proper acessibility for screen-readers.
         */
        var _modifyRequiredFieldLabels = function () {
            $(".info.required > label").each(function (e) {
                let labelHtml = $(this).html();
                $(this).html(`${labelHtml} <span class="sr-only">*</span>`);
            });
        };

        var _addApplicationLoadingIndicators = function () {
            const submitButtonSelector = ".crmEntityFormView .btn.btn-primary.submit-btn";

            const observer = new MutationObserver(_displayLoadingIndicator);
            const config = { attributes: true };
            const targetNodes = document.querySelectorAll(submitButtonSelector);

            targetNodes.forEach((node) => observer.observe(node, config));
        };

        var _displayLoadingIndicator = function (mutationsList) {
            mutationsList.forEach((mutation) => {
                const disabled = mutation.target.disabled;
                if (disabled && !mutation.target.classList.contains("read-only-enabled")) {
                    FRCC?.Portal?.Utilities?.showProgressIndicator("Saving...");
                    return;
                }

                FRCC?.Portal?.Utilities?.hideProgressIndicator();
            });
        };

        var _modifyNativeModalIFrames = function () {
            $("iframe[data-page]").on("load", function (e) {
                const iFrame = $(e.target);
                const iFrameDocument = iFrame[0]?.contentWindow?.document;
                const iFrame$ = e?.currentTarget?.contentWindow?.$;
                if (!iFrame || iFrame.length === 0 || !iFrameDocument || !iFrame$) {
                    return;
                }

                _addRequiredLegend(iFrame$);
                _updateModalIFrameHeight(iFrame, iFrame$, true);
                _registerMutationObserver(
                    iFrameDocument,
                    "body",
                    true,
                    { attributes: true, childList: true, subtree: true, attributeFilter: ["style", "class"] },
                    () => _updateModalIFrameHeight(iFrame, iFrame$)
                );
                _registerResizeObserver(iFrameDocument, "td.cell", true, () => _updateModalIFrameHeight(iFrame, iFrame$));
                _registerResizeObserver(iFrameDocument, "#ValidationSummaryEntityFormControl_EntityFormView li", true, () =>
                    _updateModalIFrameHeight(iFrame, iFrame$)
                );

                iFrame$("[data-toggle=tooltip]").tooltip();
                iFrame$("[data-toggle=tooltip]").tooltip({
                    container: ".input-group",
                    position: "top",
                });
                
                // Adding code to show spinner on submit button
                const modalSubmitButtonSelector = ".crmEntityFormView .submit-btn.btn.btn-primary.form-action-container-left";
                const modalSubmitButtons = iFrameDocument.querySelectorAll(modalSubmitButtonSelector);

                if (modalSubmitButtons.length > 0) {
                    const observer = new MutationObserver(_displayModalLoadingIndicator);
                    const config = { attributes: true };

                    // Attach the observer to each submit button
                    modalSubmitButtons.forEach((button) => observer.observe(button, config));
                }
            });
        };

        //new function for modal spinner above
        var _displayModalLoadingIndicator = function (mutationsList) {
            mutationsList.forEach((mutation) => {
                const disabled = mutation.target.disabled;
                if (disabled && !mutation.target.classList.contains("read-only-enabled")) {
                    // Create an accessible progress indicator
                    const progressElementId = "progress-indicator";
                    let progressElement = document.getElementById(progressElementId);
                    
                    // If the progress element doesn't exist, create it
                    if (!progressElement) {
                        progressElement = document.createElement("div");
                        progressElement.id = progressElementId;
                        progressElement.setAttribute("role", "alert"); // Announces changes to the content
                        progressElement.setAttribute("aria-live", "assertive"); // Screen readers read changes immediately
                        progressElement.style.position = "absolute";
                        progressElement.style.left = "-9999px"; // Keep it off-screen visually
                        document.body.appendChild(progressElement);
                    }
        
                    // Update the progress indicator content
                    progressElement.textContent = "Saving...";
                    
                    FRCC?.Portal?.Utilities?.showProgressIndicator("Saving...");
                    
                    // Add a 3-second timer
                    setTimeout(() => {
                        FRCC?.Portal?.Utilities?.hideProgressIndicator();
                        progressElement.textContent = ""; // Clear the content after the action
                    }, 3000); // 3000 milliseconds = 3 seconds
        
                    return;
                }
            });
        };

        var _addRequiredLegend = function ($Frame) {
            const $requiredLegend = $Frame(`<p class="control-label block required-legend text-right mb-4">Required</p>`);
            $Frame(".crmEntityFormView").parent().prepend($requiredLegend);
        };

        var _updateModalIFrameHeight = function (iFrame, iFrame$, onLoad = false) {
            const $body = iFrame$("body");
            const scrollHeight = $body.prop("scrollHeight"); // To determine if IFrame is unloaded, check scrollHeight
            if (scrollHeight === 0) { // If unloaded, exit
                return;
            }

            const controlHeight = iFrame$("#EntityFormControl").height(); // Retrieve EntityFormControl element's height

            const lookupButtons = iFrame$.find("button.btn.btn-default.launchentitylookup > span.fa.fa-search");
            if (lookupButtons.length && controlHeight < MIN_LOOKUP_IFRAME_HEIGHT) {
                iFrame.height(MIN_LOOKUP_IFRAME_HEIGHT);
            } else {
                iFrame.height(controlHeight);
            }

            $body.height(controlHeight);

            if (onLoad) {
                $("html").scrollTop(0);
            }
        };

        var handleError = function (errorMessage, displayError = false) {
            if (displayError) {
                alert(errorMessage);
            }

            console.error(errorMessage);
        };

        /**
         * Parent function that contains a set of functions aimed at adjusting the
         * multistep (advanced form) navigation elements and functionality.
         *
         * @async
         * @function _modifyMultiStepNavigationElements
         */
        const _modifyMultiStepNavigationElements = async function () {
            // Note: Ordering of the functions is important
            _modifyMultiStepListLayout();
            _replaceMultistepFormNavigationHtml();
            await _modifyMultistepFormNavigation();
            _modifyMultiStepListGroupProgressSteps();
        };

        /**
         * Adjusts the markup to handle the two column layout for Multi-Step Form Layouts
         */
        var _modifyMultiStepListLayout = function () {
            var multiStepList = document.querySelector("ol.progress.list-group");

            if (multiStepList) {
                var $multiStepLayoutContainer = $(multiStepList).parent().parent();
                var $multiStepFormContent = $("#WebFormPanel");

                $multiStepLayoutContainer.addClass("multi-step-form-layout");
                $multiStepFormContent.wrap('<div class="multi-step-form-main-content"></div>');
                $(multiStepList).closest("div").wrap('<nav class="multi-step-form-list-group" aria-label="application steps"></nav>');
                $(multiStepList).removeClass("top");
            }
        };

        /**
         * Adjusts the multistep form list group progress to include screen reader text
         */
        var _modifyMultiStepListGroupProgressSteps = function () {
            var multiStepList = document.querySelector("div.progress.list-group");

            if (multiStepList) {
                var listItems = multiStepList.querySelectorAll("a");

                // Hide extra markup not needed for screen readers
                var $multiStepProgressDescription = $(multiStepList).next(".sr-only");
                if ($multiStepProgressDescription.length) {
                    $multiStepProgressDescription.remove();
                }

                // Add screen reader text to each list item
                listItems.forEach(function (item) {
                    var textContent = "";
                    var listItemSRtext = document.createElement("span");
                    listItemSRtext.classList.add("sr-only");

                    if (item.classList.contains("list-group-item-success")) {
                        item.querySelector(".glyphicon").removeAttribute("aria-label");
                        item.querySelector(".glyphicon").removeAttribute("role");
                        textContent = "Completed";
                    } else if (item.classList.contains("active")) {
                        textContent = "Current";
                    } else if (item.classList.contains("incomplete")) {
                        textContent = "Not Completed";
                    }

                    listItemSRtext.textContent = textContent;
                    item.appendChild(listItemSRtext);
                });
            }
        };

        /**
         * Utility function to help make disabled elements more accessible with readonly attributes
         * and disabling unselected values so values cant be changed
         */
        var updateDisabledElements = function (item) {
            item.setAttribute("readonly", "");

            item.removeAttribute("disabled");
            item.removeAttribute("aria-disabled");
        };

        /**
         * Disabled radio buttons cant be read by screen readers or recieve focus,
         * we will add readonly attributes to the :checked:disabled options and remove the disabled attribute
         * Just changing radio buttons to readonly doesnt prevent the value from changing
         */
        var _removeRadioButtonDisabledAttribute = function () {
            var disabledCheckedRadios = document.querySelectorAll('input[type="radio"]:checked:disabled');

            disabledCheckedRadios.forEach((item) => {
                updateDisabledElements(item);
            });
        };

        /**
         * Remove disabled attributes from input elements that are not radios or checkboxes
         */
        var _removeInputDisabledAttribute = function () {
            var disabledInputs = document.querySelectorAll('input:not([type="radio"]):not([type="checkbox"]):disabled');

            disabledInputs.forEach((item) => {
                updateDisabledElements(item);
            });
        };

        /**
         * Remove disabled attributes from select elements, apply disabled to options so value cant be changed
         */
        var _removeSelectDisabledAttribute = function () {
            var disabledSelect = document.querySelectorAll("select:disabled");

            disabledSelect.forEach((item) => {
                updateDisabledElements(item);

                var unselectedOptions = item.querySelectorAll("option:not(:checked)");
                unselectedOptions.forEach((item) => {
                    item.setAttribute("disabled", "");
                });
            });
        };

        /**
         * Remove disabled attributes from textarea elements
         */
        var _removeTextareaDisabledAttribute = function () {
            var disabledTextareas = document.querySelectorAll("textarea:disabled");

            disabledTextareas.forEach((item) => {
                updateDisabledElements(item);
            });
        };

        /**
         * Function to call all the form related functions that removed disabled attributes and adds readonly, disables non selection options if necessary so values cannot be changed
         */
        var _removeFormElementDisabledAttributes = function () {
            _removeRadioButtonDisabledAttribute();
            _removeInputDisabledAttribute();
            _removeSelectDisabledAttribute();
            _removeTextareaDisabledAttribute();
        };

        /**
         * This function finds links in description elements and hten will set target attr to _new and
         * inject whatever code is stored in AccessibilityHelpText content snippet
         */
        var _insertHelpSpan = function () {
            $("main.page-content.page-application a").each(async function () {
                $(this).attr("target", "_blank");
                const contentSnippetResponse = await FRCC.Portal.Utilities.asyncAppAjax({
                    type: "GET",
                    url: `/Services/GetContentSnippet?contentSnippetName=EYGA/Accessibility/Portal/NewTab`,
                    contentType: "application/json",
                }, false);
                const contentSnippet = JSON.parse(contentSnippetResponse).results[0];
                $(this).append(contentSnippet.Value);
            });
        };

        const _modifyMultistepFormNavigation = async function () {
            const applicationId = $("#hdn-parentrecordid").val();
            if (!applicationId) {
                $(".multi-step-form-list-group div a:not(.active)").addClass("disabled");
                return;
            }

            try {
                const webFormSessionResponse = await FRCC.Portal.Utilities.asyncAppAjax({
                    type: "GET",
                    url: `/Services/GetWebFormSessions/?applicationId=${applicationId}`,
                    contentType: "application/json",
                });

                const webFormSession = JSON.parse(webFormSessionResponse);
                if (!webFormSession || !webFormSession.Id) {
                    return;
                }

                $(".multi-step-form-list-group").attr("data-sessionid", webFormSession.Id);

                const webFormSteps = webFormSession.WebFormSteps;
                const orderedWebFormSteps = _orderWebFormSteps(webFormSteps);

                await _attachMultistepFormNavigationHandlers(webFormSession.StepHistory, orderedWebFormSteps, webFormSession.Id);
            } catch (error) {
                console.error(error);
            }
        };

        const _replaceMultistepFormNavigationHtml = function () {
            const $ol = $(".multi-step-form-list-group ol");
            const $newDiv = $("<div>", {
                class: $ol.attr("class"),
                id: $ol.attr("id"),
            });

            $ol.find("li").each(function () {
                const $li = $(this);
                const $a = $("<a>", {
                    class: $li.attr("class"),
                    html: $li.html(),
                });

                $newDiv.append($a);
            });

            $ol.replaceWith($newDiv);
        };

        const _attachMultistepFormNavigationHandlers = async function (webFormSessionStepHistory, orderedWebFormSteps, webFormSessionId) {
            const visitedIds = new Set(webFormSessionStepHistory.map((step) => step.ID));
            const updatedWebFormSteps = orderedWebFormSteps.map((step) => {
                return {
                    ...step,
                    Visited: visitedIds.has(step.Id),
                };
            });

            $(".multi-step-form-list-group div a").each(async function (index) {
                const $webFormStepLink = $(this);

                const matchedWebFormStep = updatedWebFormSteps.filter((s) => s.Order === index)[0];
                $webFormStepLink.attr({
                    "data-visited": matchedWebFormStep?.Visited,
                    "data-stepid": matchedWebFormStep?.Id,
                    "data-steporder": matchedWebFormStep?.Order
                });

                const addCheckIcon = () => $webFormStepLink.prepend(`<span class="glyphicon glyphicon-ok"></span>`);

                if ($webFormStepLink.hasClass("active")) {
                    const isReadonly = document.getElementById("application-readonly-alert");

                    // Show check on step if app readonly
                    if (matchedWebFormStep.Visited && isReadonly) {
                        addCheckIcon();
                    }

                    return;
                }

                if (!matchedWebFormStep || !matchedWebFormStep.Visited) {
                    $webFormStepLink.addClass("disabled");
                    return;
                }

                $webFormStepLink.attr("tabindex", "0");
                $webFormStepLink.attr("href", "javascript:void(0)");

                if (!$webFormStepLink.hasClass("list-group-item-success")) {
                    $webFormStepLink.addClass("list-group-item-success").removeClass("incomplete");
                    addCheckIcon();
                }

                $(this).click(async function (e) {
                    await _updateWebFormSession(e, matchedWebFormStep, webFormSessionId);
                });
            });

            /* For Download View forms, we want to want to tag which sections have been
             * visited, so we know which Yes / No fields to clear. */
            var prevEntityPanel = null;
            $(".accordion-container #EntityFormPanel").each(function (index) {
                const matchedWebFormStep = updatedWebFormSteps.filter((s) => s.Order === index)[0];

                $(this).attr("data-completed", false);
                $(this).attr("data-visited", matchedWebFormStep.Visited);
                if (prevEntityPanel && matchedWebFormStep.Visited) {
                    prevEntityPanel.attr("data-completed", true);
                }
                prevEntityPanel = $(this);
            });
        };

        const _updateWebFormSession = async function (e, webFormStep, webFormSessionId) {
            const webFormStepId = webFormStep?.Id;
            if (!webFormStepId) {
                return;
            }

            const confirm = await confirmUnsavedChanges();
            if (!confirm) {
                return;
            }

            await updateWebFormSessionHelper(webFormStep, webFormSessionId);
        };

        const updateWebFormSessionHelper = async function (webFormStep, webFormSessionId) {
            const webFormStepId = webFormStep?.Id;
            const webFormStepName = webFormStep.Name;

            FRCC.Portal.Utilities.showProgressIndicator(`Navigating to Form Step ${webFormStepName}`);

            const updateWebFormSessionResponse = await FRCC.Portal.Utilities.asyncAppAjax({
                type: "POST",
                url: `/_api/eyfrcc_updatewebformsessionrequests`,
                data: JSON.stringify({
                    ["eyfrcc_webformsessionid"]: webFormSessionId,
                    ["eyfrcc_webformstepid"]: webFormStepId,
                    ["eyfrcc_webformstepindex"]: webFormStep.Order,
                }),
            });

            window.location.reload();
        };

        const _orderWebFormSteps = function (webFormSteps) {
            const stepMap = {};
            webFormSteps.forEach((step) => {
                stepMap[step.Id] = step;
            });

            // Find the starting step (a step that isn't anyone else's `NextStepId`)
            const nextStepIds = new Set(webFormSteps.map((step) => step.NextStepId));
            const startingStep = webFormSteps.find((step) => !nextStepIds.has(step.Id));

            // Reorder the steps based on the NextStepId
            const orderedSteps = [];
            let currentStep = startingStep;
            let order = 0;

            while (currentStep) {
                orderedSteps.push({ ...currentStep, Order: order });
                order++;

                const nextId = currentStep.NextStepId;
                currentStep = stepMap[nextId] || null;
            }

            return orderedSteps;
        };

        /**
         * Hide the instructions content if there isnt anything written for the instructions text
         */
        var _hideInstructionsIfEmpty = function () {
            const instructionsBlock = document.querySelector('[data-id="instructions-block"]');

            if (instructionsBlock) {
                const instructionsContent = instructionsBlock.querySelector(".override-rich-text");

                if (instructionsContent.innerText === "") {
                    instructionsBlock.style.display = "none";
                }
            }
        };

        /**
         * Reverse the order of the boolean yes/no cells so 'Yes' is first,
         * Also deselect the 'No' option if the next step isnt visited yet as
         * a way to determine if those choices were done by the user since they
         * would have chosen them to get to the following step
         */
        var _reverseRadioBooleanCells = function () {
            $("iframe[data-page]").on("load", function (e) {
                try {
                    _reverseRadioBooleanCellsHelper(e.currentTarget.contentWindow.document);
                } catch (error) {
                    handleError(error);
                }
            });

            _reverseRadioBooleanCellsHelper(document);
        };

        var _reverseRadioBooleanCellsHelper = function (documentContext) {
            const yesNoCells = documentContext.querySelectorAll(".boolean-radio-cell .boolean-radio");

            if (yesNoCells.length) {
                yesNoCells.forEach((cell) => {
                    const disabledRadioSet = cell.querySelectorAll(".aspNetDisabled");
                    let noOptionInput;
                    let yesOptionInput;
                    let yesOptionLabel;

                    // If the radios are readonly we need to get the parent spans separating the two options,
                    // markup is different between readonly/disabled an enabled
                    if (disabledRadioSet.length) {
                        noOptionInput = cell.querySelectorAll("input")[0].closest("span.aspNetDisabled");
                        yesOptionInput = cell.querySelectorAll("input")[1].closest(".aspNetDisabled");
                        yesOptionLabel = cell.querySelectorAll("label")[1].closest(".aspNetDisabled");
                    } else {
                        noOptionInput = cell.querySelectorAll("input")[0];
                        yesOptionInput = cell.querySelectorAll("input")[1];
                        yesOptionLabel = cell.querySelectorAll("label")[1];
                    }

                    cell.insertBefore(yesOptionInput, noOptionInput);
                    cell.insertBefore(yesOptionLabel, noOptionInput);
                });
            }
        };

        // Function to trigger the unchecking in the edit form
        var _checkForEditElementAndUncheckRadios = function () {
            const $editIFrame = $(".modal-form-edit iframe[data-page]");

            $editIFrame.on("load", function (e) {
                try {
                    const iframeDoc = e.currentTarget.contentWindow.document;

                    if (iframeDoc.getElementById("eyfrcc_continuationsubapplicantapplicationallowed")) {
                        _uncheckDefaultedRadioButtonsHelperForEditForm(iframeDoc);
                    }
                } catch (error) {
                    handleError(error);
                }
            });
        };

        var _uncheckDefaultedRadioButtonsHelperForEditForm = function (documentContext) {
            const yesNoCells = documentContext.querySelectorAll(".boolean-radio-cell .boolean-radio");

            // Only proceed if there are radio buttons to work with
            if (yesNoCells.length) {
                yesNoCells.forEach((cell) => {
                    if (!cell.hasAttribute("disabled")) {
                        // Uncheck both 'Yes' and 'No' options
                        const yesRadio = cell.querySelector('input[value="1"]');
                        const noRadio = cell.querySelector('input[value="0"]');
                        if (yesRadio) {
                            yesRadio.checked = false;
                        }
                        if (noRadio) {
                            noRadio.checked = false;
                        }
                    }
                });
            }
        };

        /**
         * Deselect the 'No' option if the next step isnt visited yet as
         * a way to determine if those choices were done by the user since they
         * would have chosen them to get to the following step
         */
        var _uncheckDefaultedRadioButtons = function () {
            const $insertIFrame = $(".modal-form-insert iframe[data-page]");

            $insertIFrame.on("load", function (e) {
                try {
                    _uncheckDefaultedRadioButtonsHelper(e.currentTarget.contentWindow.document, true);
                } catch (error) {
                    handleError(error);
                }
            });

            _uncheckDefaultedRadioButtonsHelper(document);
        };

        var _uncheckDefaultedRadioButtonsHelper = function (documentContext, isIFrameInsertForm = false) {
            const currentStep = documentContext.querySelector(".list-group-item.active");

            const yesNoCells = documentContext.querySelectorAll(".boolean-radio-cell .boolean-radio");

            if (isIFrameInsertForm) {
                if (yesNoCells.length) {
                    yesNoCells.forEach((cell) => {
                        if (!cell.hasAttribute("disabled")) {
                            // input[value="0"] is the 'No' option
                            cell.querySelector('input[value="0"]').checked = false;
                        }
                    });
                }
            }

            if (!!currentStep === true) {
                const isNextStepCompleted = currentStep.nextElementSibling?.dataset.visited === "true";
                
                let fieldMappingsConfiguration = window.fieldMappingsConfiguration || {};
                
                if (yesNoCells.length) {
                    yesNoCells.forEach((cell) => {
                        if (!isNextStepCompleted && !cell.hasAttribute("disabled") && !fieldMappingsConfiguration?.FieldMappings?.includes(cell.id)) {
                            // input[value="0"] is the 'No' option
                            cell.querySelector('input[value="0"]').checked = false;
                        }
                    });
                }
            }

            /* For the Download View page, we will instead query for and uncheck all
             * Yes / No fields in any basic forms that have been tagged as not yet visited */
            documentContext.querySelectorAll("#EntityFormPanel[data-completed=false] .boolean-radio-cell .boolean-radio").forEach((cell) => {
                cell.querySelector('input[value="0"]').checked = false;
            });
        };

        const _removeAriaLabels = function () {
            $(".crmEntityForm [aria-label]").each(function () {
                const $element = $(this);
                const ariaLabel = $element.attr("aria-label")?.toLowerCase();
                const tagName = this.tagName.toLowerCase();

                const isActionMenuButton = ariaLabel === "action menu" && tagName === "button";
                const isCloseButton = ariaLabel === "close";
                const hasRequiredClass = $element.hasClass("multi-step-form-list-group") || $element.hasClass("modal");

                if (!(isActionMenuButton || isCloseButton || hasRequiredClass)) {
                    $element.removeAttr("aria-label");
                }
            });
        };

        var _removeAriaReadOnlyOnTables = function () {
            const viewGridDivs = document.querySelectorAll("div.view-grid.table-responsive");
            viewGridDivs.forEach((div) => {
                const ariaReadonlyElements = div.querySelectorAll("[aria-readonly]");
                ariaReadonlyElements.forEach((element) => {
                    element.removeAttribute("aria-readonly");
                });
            });
        };

        var _applyDatePickerAriaAttributes = function () {
            $("iframe[data-page]").on("load", function (e) {
                try {
                    _applyDatePickerAriaAttributesHelper(e.currentTarget.contentWindow.document);
                } catch (error) {
                    handleError(error);
                }
            });

            _applyDatePickerAriaAttributesHelper(document);
        };

        var _applyDatePickerAriaAttributesHelper = function (documentContext) {
            const datePickerEls = documentContext.getElementsByClassName("datetimepicker");

            if (!!datePickerEls.length) {
                datePickerEls.forEach((datepicker) => {
                    const datePickerInput = datepicker.querySelector("input");
                    const datePickerButton = datepicker.querySelector('span[role="button"]');

                    datePickerButton.setAttribute("aria-controls", `${datePickerInput.id}-datetimepicker`);

                    // Mutation observer for date widget showing
                    const config = {
                        attributes: true,
                        attributeFilter: ["aria-expanded"],
                    };

                    const callback = function (mutationsList, observer) {
                        for (const mutation of mutationsList) {
                            if (mutation.type === "attributes" && mutation.attributeName === "aria-expanded") {
                                const isAriaExpanded = datePickerButton.getAttribute("aria-expanded") === "true";
                                if (isAriaExpanded) {
                                    datepicker.querySelector(".bootstrap-datetimepicker-widget").setAttribute("id", `${datePickerInput.id}-datetimepicker`);
                                }
                            }
                        }
                    };

                    const observer = new MutationObserver(callback);
                    observer.observe(datePickerButton, config);
                });
            }
        };

        var _addTableCaptions = function () {
            $("iframe[data-page]").on("load", function (e) {
                try {
                    _addTableCaptionsHelper(e.currentTarget.contentWindow.document);
                } catch (error) {
                    handleError(error);
                }
            });

            _addTableCaptionsHelper(document);
        };

        var _addTableCaptionsHelper = function (documentContext) {
            const tableSubgrids = documentContext.querySelectorAll(".cell.subgrid-cell");

            tableSubgrids.forEach((tableSubgrid) => {
                const tableEl = tableSubgrid.querySelector("table");
                const tableLabel = tableSubgrid.querySelector("h3 .field-label");
                const captionEl = document.createElement("caption");

                captionEl.classList.add("sr-only");
                captionEl.append(tableLabel ? tableLabel.innerText : "Table");

                tableEl.prepend(captionEl);
            });
        };

        var _addDocumentFileAriaLabels = function (mutationsList) {
            for (const mutation of mutationsList) {
                if (mutation.type === "childList") {
                    mutation.addedNodes.forEach((node) => {
                        if (node.nodeType === Node.ELEMENT_NODE && node.tagName === "TR") {
                            const fileButton = node.querySelector("td:first-child button");
                            const fileName = fileButton?.innerText;

                            fileButton.setAttribute("aria-label", `${fileName}, download document`);
                        }
                    });
                }
            }
        };

        var _addDocumentDownloadAccessibleText = function () {
            const documentFilesTable = document.getElementById("file-upload-table");

            if (documentFilesTable) {
                const observer = new MutationObserver(_addDocumentFileAriaLabels);

                const config = { childList: true, subtree: true };
                observer.observe(documentFilesTable, config);
            }
        };

        var _fixTableActionsHeader = function () {
            $("iframe[data-page]").on("load", function (e) {
                try {
                    _fixTableActionsHeaderHelper(e.currentTarget.contentWindow.document);
                } catch (error) {
                    handleError(error);
                }
            });

            _fixTableActionsHeaderHelper(document);
        };

        var _fixTableActionsHeaderHelper = function (documentContext) {
            // Find TH and remove sr-only class
            const tableActionsTh = documentContext.querySelectorAll(`th[data-th="<span class='sr-only'>Actions</span>"]`);

            tableActionsTh.forEach((thElement) => {
                const srElement = thElement.querySelector(".sr-only");
                srElement?.classList.remove("sr-only");
            });
        };

        /**
         * The table class in application forms is preventing the table from being usable for mobile
         * Adjust parent table class names and get rid of striped table class
         *
         */
        var _updateTableResponsive = function () {
            const tableResponsiveElements = document.querySelectorAll(`.table-responsive`);

            tableResponsiveElements.forEach((tableResponsive) => {
                const table = tableResponsive.querySelector("table");

                tableResponsive.classList.remove("table-responsive");
                tableResponsive.classList.add("table-scroll-wrapper");

                table.className = "table table-custom table-mobile table-fluid-label-width-md";
            });
        };

        /**
         * The Action buttons in the application tables use the wrong icon and need
         * extra classes to help display it correctly on mobile
         * Use this utility function to take the dropdown button and create the new icon and sr-only text
         * and also apply parent td classes to help with mobile layout
         *
         */
        var _setUpActionButton = function (button) {
            const dropdownTd = button.closest("td");
            const dropdownButtonIcon = button.querySelector(".fa");
            const mobileActionsElement = document.createElement("span");
            const mobileActionsCaret = document.createElement("span");

            mobileActionsElement.className = "visible-xs";
            mobileActionsElement.innerText = "Actions";

            mobileActionsCaret.setAttribute("aria-hidden", "true");
            mobileActionsCaret.className = "caret ml-2";

            mobileActionsElement.append(mobileActionsCaret);
            button.append(mobileActionsElement);

            dropdownTd.classList.add("contains-component", "text-right");

            button.className = "btn btn-default btn-xs-breakpoint-md btn-xs-breakpoint-lg";

            if (dropdownButtonIcon) {
                dropdownButtonIcon.className = "glyphicon glyphicon-option-horizontal hidden-xs";
            }
        };

        /**
         * Update the tables action buttons when they intitially load, when new elements are added via create or delete
         * the icons were reverting back to incorrect icon, so there is a mutation observer to listen for changes in the table
         * in the `_updateTableActionIcons` function
         *
         */
        var _updateInitialTableActions = function () {
            // Find action buttons and update classes and icon
            const tableActionsDropdown = document.querySelectorAll(`td .dropdown`);

            tableActionsDropdown.forEach((dropdown) => {
                const dropdownButton = dropdown.querySelector("button");
                _setUpActionButton(dropdownButton);
            });
        };

        /**
         *
         * Mutation observer function for changes in application table
         */
        var _updateNewTableActions = function (mutationsList) {
            for (const mutation of mutationsList) {
                if (mutation.type === "childList") {
                    mutation.addedNodes.forEach((node) => {
                        if (node.nodeType === Node.ELEMENT_NODE) {
                            const dropdownButton = node.querySelectorAll(".dropdown button");
                            dropdownButton.forEach((button) => {
                                _setUpActionButton(button);
                            });
                        }
                    });
                }
            }
        };

        /**
         * Listen to application table changes to update action button icons
         */
        var _updateTableActionIcons = function () {
            const tableGrids = document.querySelectorAll(".view-grid table");
            if (tableGrids.length) {
                // Update action buttons on load
                _updateInitialTableActions();

                // Update action buttons when they are added to table as new rows
                const observer = new MutationObserver(_updateNewTableActions);
                const config = { childList: true, subtree: true };

                tableGrids.forEach((table) => {
                    observer.observe(table, config);
                });
            }
        };

        /**
         * This is an accessbility-specific function that will set lookup tooltips
         * to be placed on top instead of the default of to the right (tooltip on the right get cut off by small screens)
         */
        var setTooltipPlacement = function () {
            $(document).ready(function () {
                $("[data-toggle=tooltip]").tooltip({
                    container: "body",
                    position: "top",
                });
            });
        };

        const _applyCurrencyFieldPostProcessing = function (isReadonly) {
            $("iframe[data-page]").on("load", function (e) {
                try {
                    _applyCurrencyFieldPostProcessingHelper(e.currentTarget.contentWindow.$, isReadonly);
                } catch (error) {
                    handleError(error);
                }
            });

            _applyCurrencyFieldPostProcessingHelper($, isReadonly);
        };

        const _applyCurrencyFieldPostProcessingHelper = function ($Frame, isReadonly) {
            if (!$Frame) {
                return;
            }

            const $moneyControls = $Frame(".money.form-control-cell > .control > div");
            $moneyControls.each(function () {
                const $moneyControl = $Frame(this);
                _renderCurrencyFieldSymbol($moneyControl, isReadonly);
            });
        };

        const _renderCurrencyFieldSymbol = function ($moneyControl, isReadonly) {
            if (!$moneyControl) {
                return;
            }

            // Add the $ Symbol if it is not already present
            if ($moneyControl.hasClass("input-group")) {
                return;
            }

            $moneyControl.addClass("input-group");
            
            // do not add $ if the money field is readonly with emdash (no value)
            let moneyInputEmdash = $moneyControl.find("div.text-muted");
            let emdashIsPresent = moneyInputEmdash && moneyInputEmdash.length > 0;
            
            let moneyInput = $moneyControl.find("input");
            let moneyValueIsEmpty = !moneyInput.value;
            
            if ((isReadonly && moneyValueIsEmpty) || emdashIsPresent) {
                // do nothing - we do not want a $ with our emdash
            } else {
                $moneyControl.prepend(`<span class="input-group-addon">$</span>`);
            }
        };

        const _checkForUnsavedChangesPrevious = function () {
            const $previousButton = $("#PreviousButton");
            if (!$previousButton.length) {
                return;
            }

            const previousFunction = $previousButton.prop("onclick"); // Cache button's original onclick event

            $previousButton
                .removeAttr("onclick")
                .off("click")
                .on("click", async function (event) {
                    const confirm = await confirmUnsavedChanges();
                    if (!confirm) {
                        return;
                    }

                    previousFunction(event); // Invoke button's original onclick event
                });
        };

        const confirmUnsavedChanges = async function () {
            const $dirty = $(".dirty").not("[id*='PcfControl_']"); // Exclude PCF Controls for now - native issue where they are initialized with .dirty
            if (!$dirty.length) {
                return true;
            }

            const configuration = {
                Message: `You have unsaved changes in this section. Please click Cancel then Next to save those changes, or click OK to navigate without saving.`,
                Title: "Unsaved Changes",
                CancelButtonLabel: "Cancel",
                ConfirmButtonLabel: "Ok",
                ShowCancelButton: true,
                ShowConfirmationButton: true,
                NotificationLevel: "Warning",
                ModalSize: "Medium",
            };

            const confirmationResult = await FRCC.AmeriCorps.Modals.Confirmation.openModalWithPromises(configuration);
            return confirmationResult?.Success;
        };

        const _fixNativeNotificationHTMLAccessibility = function () {
            // This could potentially be refactored to use _registerMutationObserver function; wasn't aware of that when writing
            const target = document.getElementById("liquid_form");
            const config = { attributes: false, childList: true, subtree: true };
            const mutationCallback = (mutationList, observer) => {
                $('.notification:not(".accessibility-fix")').each((index, notificationDiv) => {
                    let $notificationDiv = $(notificationDiv);

                    // add a class to this notification so we only process it once
                    $notificationDiv.addClass("accessibility-fix");

                    // check if notification has button before the other content - this is an accessibility violation
                    let divContents = $notificationDiv.contents();
                    if (divContents.length > 0 && divContents[0].type === "button") {
                        // remove and re-add the button to the end of the content
                        let $button = $notificationDiv.children("button");
                        $button.remove();
                        $notificationDiv.append($button);
                    }
                });
            };

            const mutationObserver = new MutationObserver(mutationCallback);
            mutationObserver.observe(target, config);
        };

        const _removePlaceholderText = function () {
            var targetSelector = document.querySelector("iframe.quickform");
            if (targetSelector) {
                var observer = new MutationObserver(function (mutations) {
                    mutations.forEach(function (mutation) {
                        _removePlaceholderTextHelper();
                    });
                });
                observer.observe(targetSelector, { attributes: true, childList: true, characterData: true, subtree: true });

            }
            _removePlaceholderTextHelper();
        }

        const _removePlaceholderTextHelper = function () {
            //Remove telephone from iframe
            var targetSelector = document.querySelector("iframe.quickform");
            var targetDoc = targetSelector ? targetSelector.contentWindow.document : document;
            let phoneInputs = targetDoc.querySelectorAll("input[placeholder='Provide a telephone number']");
            phoneInputs.forEach(inputControl => {
                _replaceWithEmdash(inputControl);
            });

            //Remove Datepicker from document
            let dateInputs = document.querySelectorAll("input[placeholder='M/D/YYYY']");
            dateInputs.forEach(inputControl => {
                _replaceWithEmdash(inputControl);
            });

            let selectOptions = document.querySelectorAll("option[selected='Selected'][label='Select']");
            selectOptions.forEach(selectControl => {
                selectControl.removeAttribute("label");
                selectControl.textContent = '';
                
            });
        }

        const _replaceWithEmdash = function (inputControl) {
            if (inputControl) {
                inputControl.removeAttribute("placeholder");
                if (inputControl.value === null || inputControl.value === "") {
                    let parentElement = inputControl.parentElement;
                    parentElement.setAttribute("style", "position: relative;")
                    parentElement.insertAdjacentHTML('beforeend', '<div class="text-muted" aria-label="No data" style="position: absolute; top: 4px;">—</div>');
                }
            }
        }

        /**
         * Ensure that containers that contain errors are properly configured for accessibility.
         * - Add error role to the div
         * - Ensure if there is a list of items inside, there is a header
         * - Ensure if there is a list of items inside, the styling of the list is improved
         */
        const _fixErrorAlertLists = function () {
            $("iframe[data-page]").on("load", function (e) {
                try {
                    _fixErrorAlertListsHelper(e.currentTarget.contentWindow.$);
                } catch (error) {
                    handleError(error);
                }
            });

            _fixErrorAlertListsHelper($);
        }

        const _fixErrorAlertListsHelper = function ($) {
            if (!$) {
                return;
            }
            
            let errorContainers = $("div.alert.alert-danger");
            errorContainers.attr("role", "alert");
            
            errorContainers.each((index, element) => {
                let $element = $(element);
                if (!$element.hasClass("no-header")) {
                    // Add validation header if missing
                    let validationHeader = $element.find("h2.validation-header");
                    if (!validationHeader || validationHeader.length === 0) {
                        $element.prepend(
                            `<h2 class='validation-header' role='none'>
                                <span role='presentation' class='fa fa-info-circle'></span>
                                Please correct the following errors:
                            </h2>`);
                    }
                }
            });
        }

         /**
          * - Replace <label>s under fieldset[role='radiogroup'] with <legend>s
          * - Replace <label>s under .boolean-radio-cell with <legend>s
          * - Update "Required *" legend on all pages to use required-legend class and remove unneeded <span>
          * - Update OOB controls to remove redundant span.sr-only that duplicate the field-label
         */ 
         const _replaceApplicableLabelsWithLegends = function () {
            $("iframe[data-page]").on("load", function (e) {
                try {
                    let $frame = e.currentTarget.contentWindow.$;
                    if (!$frame) {
                        return;
                    }

                    _replaceFieldSetLabelsWithLegends($frame);
                    _replaceBooleanRadioCellLabelsWithLegends($frame);
                    _fixRequiredLegendStying($frame);
                    _removeRedundantSROnlySpans($frame);
                } catch (error) {
                    handleError(error);
                }
            });

            _replaceFieldSetLabelsWithLegends($);
            _replaceBooleanRadioCellLabelsWithLegends($);
            _fixRequiredLegendStying($);
            _removeRedundantSROnlySpans($);
        }
        
        const _replaceFieldSetLabelsWithLegends = function ($) {
            let labels = $("fieldset[role='radiogroup']").find("label.control-label");
            labels.each((index, element) => {
                let label = $(element);
                let newElement = _replaceHTMLTag(label, "legend");
                if (newElement.hasClass("required")) {
                    newElement.parent().attr("aria-required", "true");
                }
            });
        }

        const _replaceBooleanRadioCellLabelsWithLegends = function ($) {
            let labels = $(".boolean-radio-cell").find("label.field-label");
            labels.each((index, element) => {
                let label = $(element);
                if (label.hasClass("required") || label.parent().hasClass("required")) {
                    label.parents(".boolean-radio-cell").attr("aria-required", "true");
                }
                
                let newElement = _replaceHTMLTag(label, "legend");
            });
        }

        const _fixRequiredLegendStying = function ($) {
            // get the legend that says "* Required" 
            // change required class to required-legend, and remove the inner <span>*</span>
            let requiredLegendLabel = $("p.control-label.block.required.text-right.mb-4");
            requiredLegendLabel.removeClass("required");
            requiredLegendLabel.addClass("required-legend");
            
            let innerSpans = requiredLegendLabel.find("span");
            innerSpans.remove();
        }

        const _replaceHTMLTag = function ($element, newTagName, copyAttributes = true) {
            $element.wrap(`<${newTagName}>`);
            var $newElement = $element.parent();
            
            if (copyAttributes) {
                $.each($element.prop('attributes'), function() {
                    $newElement.attr(this.name,this.value);
                });
            }
            
            $element.contents().unwrap();
            return $newElement;
        }

        const _removeRedundantSROnlySpans = function($) {
            let controls = $("div.control");
            controls.each((index, element) => {
                let $element = $(element);
                let srOnlySpans = $element.find("label > span.sr-only");

                let fieldLabel = $element.parent().find(".field-label");
                srOnlySpans.each((spanIndex, spanElement) => {
                    if (spanElement.innerText === fieldLabel.text()) {
                        $(spanElement).remove();
                    }
                });
            });
        }

        const _addApplicationFormSkipLink = () => {
            $('.skip-links a[href="#skip-header-navigation"]').after(
              '<a href="#multistep-subheading">Skip to form content</a>'
            );
        };
        
        const _removeCalendarsForReadonlyDateTimePickers = function(isReadonly) {
            $("iframe[data-page]").on("load", function (e) {
                try {
                    let $frame = e.currentTarget.contentWindow.$;
                    if (!$frame) {
                        return;
                    }

                    _removeCalendarsForReadonlyDateTimePickersHelper($frame, isReadonly);
                } catch (error) {
                    handleError(error);
                }
            });

            _removeCalendarsForReadonlyDateTimePickersHelper($, isReadonly);
        }
        
        const _removeCalendarsForReadonlyDateTimePickersHelper = function ($, isReadonly) {
            if (isReadonly) {
                $("div.datetimepicker > span.input-group-addon").remove();
            }
        }

        const _setFocus = () => {
            if (window.location.href.includes("Create") || window.location.href.includes("Edit")) {
                const focusAnchor = $("<a>", {
                    href: "#",
                    id: "focus-anchor",
                    text: "",
                    "aria-hidden": "true"
                });
                $("html").prepend(focusAnchor);
                $("#focus-anchor").focus();
            }
        };

        return {
            loadGlobalAccessibilityModifications: loadGlobalAccessibilityModifications,
            loadApplicationAccessibilityModifications: loadApplicationAccessibilityModifications,
            resetPageTabbing: resetPageTabbing,
        };
    })();
})(this);
